# -*- coding: utf-8 -*-
"""Discover and lookup command plugins.

This comes from OpenStack cliff.
"""

import importlib.metadata
import logging

LOG = logging.getLogger(__name__)


class EntryPointWrapper(object):
    """Wrap up a command class already imported to make it look like a plugin."""

    def __init__(self, name, command_class):
        self.name = name
        self.command_class = command_class

    def load(self, require=False):
        return self.command_class


class CommandManager(object):
    """Discovers commands and handles lookup based on argv data.

    :param namespace: String containing the setuptools entrypoint namespace
                      for the plugins to be loaded. For example,
                      ``'cliff.formatter.list'``.
    :param convert_underscores: Whether cliff should convert underscores to
                                spaces in entry_point commands.
    """

    def __init__(self, namespace, convert_underscores=True):
        self.commands = {}
        self.namespace = namespace
        self.convert_underscores = convert_underscores
        self._load_commands()

    def _load_commands(self):
        # NOTE(jamielennox): kept for compatability.
        self.load_commands(self.namespace)

    def load_commands(self, namespace):
        """Load all the commands from an entrypoint"""
        entry_points = importlib.metadata.entry_points()
        if hasattr(entry_points, "select"):
            entry_points = entry_points.select(group=namespace)
        else:
            entry_points = entry_points.get(namespace, [])
        for ep in entry_points:
            LOG.debug("found command %r", ep.name)
            cmd_name = (
                ep.name.replace("_", " ") if self.convert_underscores else ep.name
            )
            self.commands[cmd_name] = ep
        return

    def __iter__(self):
        return iter(self.commands.items())

    def add_command(self, name, command_class):
        self.commands[name] = EntryPointWrapper(name, command_class)

    def find_command(self, argv):
        """Given an argument list, find a command and
        return the processor and any remaining arguments.
        """
        search_args = argv[:]
        name = ""
        while search_args:
            if search_args[0].startswith("-"):
                name = "%s %s" % (name, search_args[0])
                raise ValueError("Invalid command %r" % name)
            next_val = search_args.pop(0)
            name = "%s %s" % (name, next_val) if name else next_val
            if name in self.commands:
                cmd_ep = self.commands[name]
                cmd_factory = cmd_ep.load()
                return (cmd_factory, name, search_args)
        else:
            raise ValueError("Unknown command %r" % next(iter(argv), ""))



import argparse
import sys
import tempfile
from unittest.mock import MagicMock, patch

import pytest

from gearbox.commandmanager import CommandManager
from gearbox.commands.help import HelpAction
from gearbox.commands.serve import ServeCommand
from gearbox.commands.setup_app import SetupAppCommand
from gearbox.main import main


# --- Test for ServeCommand using external wsgiref ---
def test_serve_command():
    with tempfile.NamedTemporaryFile() as f:
        f.write(b"""
[server:main]
use = egg:gearbox#wsgiref
host = 127.0.0.1
port = 8080

[app:dummy_app]
use = egg:dummy_app""")
        f.flush()

        opts = argparse.Namespace(
            config_file=f.name,
            app_name="dummy_app",
            server="wsgiref",
            server_name=None,
            daemon=False,
            pid_file=None,
            reload=False,
            reload_interval=1,
            monitor_restart=False,
            set_user=None,
            set_group=None,
            args=[],
            stop_daemon=False,
            show_status=False,
        )
        server_instance = MagicMock()  # Serve instance with serve_forever auto-mocked.
        with patch(
            "wsgiref.simple_server.make_server", return_value=server_instance
        ) as mock_make_server, patch.object(
            ServeCommand, "loadapp", return_value=MagicMock()
        ):
            cmd = ServeCommand(MagicMock(), argparse.Namespace(verbose_level=0))
            cmd.take_action(opts)
            mock_make_server.assert_called_once()
            server_instance.serve_forever.assert_called_once()


# --- Test for HelpAction using real CommandManager ---
def test_help_action(capsys):
    cm = CommandManager(namespace="test")
    cm.add_command(
        "dummy",
        MagicMock(
            return_value=MagicMock(
                deprecated=False, get_description=lambda: "Dummy command description"
            )
        ),
    )
    app = MagicMock(command_manager=cm)
    parser = argparse.ArgumentParser(prog="fakeapp")
    ns = argparse.Namespace(debug=True)
    action = HelpAction(option_strings=[], dest="help", nargs=0, default=app)
    with pytest.raises(SystemExit):
        action(parser, ns, None)
    output = capsys.readouterr().out
    assert "dummy" in output
    assert "Dummy command description" in output


# --- Test for makepackage command ---
def test_makepackage(tmp_path, monkeypatch):
    # Simulate command line call to `gearbox makepackage --name TestProject`
    script = tmp_path / "dummy_script.py"
    script.write_text("")  # dummy script placeholder
    # Change working dir so that 'makepackage' creates the project dir there.
    monkeypatch.chdir(tmp_path)

    with patch.object(sys, "argv", ["gearbox", "makepackage", "--name", "TestProject"]):
        main()

    project_dir = tmp_path / "TestProject"
    assert project_dir.is_dir()


# --- Test for setup-app command ---
def test_setup_app(tmp_path):
    project_dir = tmp_path / "project"
    project_dir.mkdir()
    config_file = project_dir / "development.ini"
    config_file.write_text("[app:main]\nuse=egg:fakeegg\n")

    fake_setup_app = MagicMock(return_value=0)

    with patch.object(
        sys, "argv", ["gearbox", "setup-app", "-c", str(config_file)]
    ), patch(
        "gearbox.commands.setup_app.appconfig",
        return_value=MagicMock(
            context=MagicMock(
                entry_point_name="main",
                protocol="app",
                distribution=MagicMock(read_text=MagicMock(return_value="fakemodule")),
            )
        ),
    ), patch.object(
        SetupAppCommand,
        "_import_module",
        return_value=MagicMock(setup_app=fake_setup_app),
    ):
        main()

    fake_setup_app.assert_called_once()


# --- Test for scaffold command ---
def test_scaffold(tmp_path):
    lookup_dir = tmp_path / "scaffolds"
    lookup_dir.mkdir()
    template_file = lookup_dir / "model.template"
    template_file.write_text("class {{target.capitalize()}}:\n    pass\n")
    project_dir = tmp_path / "project"
    project_dir.mkdir()

    with patch.object(
        sys,
        "argv",
        [
            "gearbox",
            "scaffold",
            "model",
            "TestModel",
            "-l",
            str(lookup_dir),
            "-p",
            str(project_dir),
        ],
    ):
        main()

    output_file = project_dir / "TestModel"
    assert output_file.is_file()
    content = output_file.read_text()
    # Expect the template engine to substitute, e.g., "Testmodel" from {{target.capitalize()}}
    assert "class Testmodel" in content


# --- Test for patch command ---
def test_patch(tmp_path):
    test_file = tmp_path / "test.txt"
    test_file.write_text("Hello World\n")

    with patch(
        "gearbox.commands.patch.PatchCommand._walk_flat",
        return_value=[str(test_file)],
    ), patch.object(
        sys, "argv", ["gearbox", "patch", str(test_file), "World", "-r", "Gearbox"]
    ):
        main()

    content = test_file.read_text()
    assert "Gearbox" in content



from __future__ import print_function

import argparse
import inspect
import logging
import os
import sys
import warnings

import pkg_resources

from .commandmanager import CommandManager
from .commands.help import HelpAction, HelpCommand
from .utils.plugins import find_egg_info_dir

log = logging.getLogger("gearbox")


class GearBox(object):
    NAME = os.path.splitext(os.path.basename(sys.argv[0]))[0]
    LOG_DATE_FORMAT = "%H:%M:%S"
    LOG_GEARBOX_FORMAT = (
        "%(asctime)s,%(msecs)03d %(levelname)-5.5s [%(name)s] %(message)s"
    )
    DEFAULT_VERBOSE_LEVEL = 1

    def __init__(self):
        self.command_manager = CommandManager("gearbox.commands")
        self.command_manager.add_command("help", HelpCommand)
        self.parser = argparse.ArgumentParser(
            description="TurboGears2 Gearbox toolset", add_help=False
        )

        parser = self.parser
        parser.add_argument(
            "--version",
            action="version",
            version="%(prog)s {0}".format(
                pkg_resources.get_distribution("gearbox").version
            ),
        )

        verbose_group = parser.add_mutually_exclusive_group()
        verbose_group.add_argument(
            "-v",
            "--verbose",
            action="count",
            dest="verbose_level",
            default=self.DEFAULT_VERBOSE_LEVEL,
            help="Increase verbosity of output. Can be repeated.",
        )
        verbose_group.add_argument(
            "-q",
            "--quiet",
            action="store_const",
            dest="verbose_level",
            const=0,
            help="Suppress output except warnings and errors.",
        )

        parser.add_argument(
            "--log-file",
            action="store",
            default=None,
            help="Specify a file to log output. Disabled by default.",
        )

        parser.add_argument(
            "-h",
            "--help",
            action=HelpAction,
            nargs=0,
            default=self,  # tricky
            help="Show this help message and exit.",
        )

        parser.add_argument(
            "--debug",
            default=False,
            action="store_true",
            help="Show tracebacks on errors.",
        )

        parser.add_argument(
            "--relative",
            default=False,
            action="store_true",
            dest="relative_plugins",
            help="Load plugins and applications also from current path.",
        )

    def _configure_logging(self):
        if self.options.debug:
            warnings.simplefilter("default")
            try:
                logging.captureWarnings(True)
            except AttributeError:
                pass

        root_logger = logging.getLogger("")
        root_logger.setLevel(logging.INFO)

        # Set up logging to a file
        if self.options.log_file:
            file_handler = logging.FileHandler(filename=self.options.log_file)
            formatter = logging.Formatter(
                self.LOG_GEARBOX_FORMAT, datefmt=self.LOG_DATE_FORMAT
            )
            file_handler.setFormatter(formatter)
            root_logger.addHandler(file_handler)

        # Always send higher-level messages to the console via stderr
        console = logging.StreamHandler(sys.stderr)
        console_level = {
            0: logging.WARNING,
            1: logging.INFO,
            2: logging.DEBUG,
        }.get(self.options.verbose_level, logging.DEBUG)
        console.setLevel(console_level)
        formatter = logging.Formatter(
            self.LOG_GEARBOX_FORMAT, datefmt=self.LOG_DATE_FORMAT
        )
        console.setFormatter(formatter)
        root_logger.addHandler(console)

    def run(self, argv):
        """Application entry point"""
        if not argv:
            # Print help when started without a subcommand.
            argv = ["--help"]

        try:
            self.options, remainder = self.parser.parse_known_args(argv)
            self._configure_logging()

            if self.options.relative_plugins:
                curdir = os.getcwd()
                sys.path.insert(0, curdir)
                pkg_resources.working_set.add_entry(curdir)

            try:
                self._load_commands_for_current_dir()
            except pkg_resources.DistributionNotFound as e:
                try:
                    error_msg = repr(e)
                except Exception:
                    error_msg = "Unknown Error"

                log.error(
                    "Failed to load project commands with error "
                    "``%s``, have you installed your project?" % error_msg
                )

        except Exception as err:
            if hasattr(self, "options"):
                debug = self.options.debug
            else:
                debug = True

            if debug:
                log.exception(err)
            else:
                log.error(err)

            return 1

        return self._run_subcommand(remainder)

    def _run_subcommand(self, argv):
        try:
            subcommand = self.command_manager.find_command(argv)
        except ValueError as err:
            if self.options.debug:
                log.exception(err)
            else:
                log.error(err)
            return 2

        cmd_factory, cmd_name, sub_argv = subcommand
        kwargs = {}
        if (
            "cmd_name" in self._getargspec(cmd_factory)[0]
        ):  # Check to see if 'cmd_name' is in cmd_factory's args
            kwargs["cmd_name"] = cmd_name
        cmd = cmd_factory(self, self.options, **kwargs)

        try:
            full_name = " ".join([self.NAME, cmd_name])
            cmd_parser = cmd.get_parser(full_name)
            parsed_args = cmd_parser.parse_args(sub_argv)
            return cmd.run(parsed_args)
        except Exception as err:
            log.exception(err)
            return 4

    def _load_commands_for_current_dir(self):
        egg_info_dir = find_egg_info_dir(os.getcwd())
        if egg_info_dir:
            package_name = os.path.splitext(os.path.basename(egg_info_dir))[0]

            try:
                pkg_resources.require(package_name)
            except pkg_resources.DistributionNotFound as e:
                msg = "%sNot Found%s: %s (is it an installed Distribution?)"
                if str(e) != package_name:
                    raise pkg_resources.DistributionNotFound(
                        msg % (str(e) + ": ", " for", package_name)
                    )
                else:
                    raise pkg_resources.DistributionNotFound(
                        msg % ("", "", package_name)
                    )

            dist = pkg_resources.get_distribution(package_name)
            for epname, ep in dist.get_entry_map("gearbox.plugins").items():
                self.load_commands_for_package(ep.module_name)

    def load_commands_for_package(self, package_name):
        dist = pkg_resources.get_distribution(package_name)
        for epname, ep in dist.get_entry_map("gearbox.project_commands").items():
            self.command_manager.commands[epname.replace("_", " ")] = ep

    def _getargspec(self, func):
        if not hasattr(inspect, "signature"):
            return inspect.getargspec(func.__init__)
        else:  # pragma: no cover
            sig = inspect.signature(func)
            args = [
                p.name
                for p in sig.parameters.values()
                if p.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD
            ]
            varargs = [
                p.name
                for p in sig.parameters.values()
                if p.kind == inspect.Parameter.VAR_POSITIONAL
            ]
            varargs = varargs[0] if varargs else None
            varkw = [
                p.name
                for p in sig.parameters.values()
                if p.kind == inspect.Parameter.VAR_KEYWORD
            ]
            varkw = varkw[0] if varkw else None
            defaults = (
                tuple(
                    (
                        p.default
                        for p in sig.parameters.values()
                        if p.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD
                        and p.default is not p.empty
                    )
                )
                or None
            )
            return args, varargs, varkw, defaults


def main():
    args = sys.argv[1:]
    gearbox = GearBox()
    return gearbox.run(args)



# (c) 2005 Ian Bicking and contributors; written for Paste
# (http://pythonpaste.org) Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license.php

import os
import sys


def native_(s, encoding="latin-1", errors="strict"):
    """If ``s`` is an instance of ``str``, return
    ``s``, otherwise return ``str(s, encoding, errors)``"""
    if isinstance(s, str):
        return s
    return str(s, encoding, errors)


fsenc = sys.getfilesystemencoding()


def copy_dir(
    source,
    dest,
    vars,
    verbosity=1,
    simulate=False,
    indent=0,
    sub_vars=True,
    interactive=False,
    overwrite=True,
    template_renderer=None,
    out_=sys.stdout,
):
    """
    Copies the ``source`` directory to the ``dest`` directory.

    ``vars``: A dictionary of variables to use in any substitutions.

    ``verbosity``: Higher numbers will show more about what is happening.

    ``simulate``: If true, then don't actually *do* anything.

    ``indent``: Indent any messages by this amount.

    ``sub_vars``: If true, variables in ``_tmpl`` files and ``+var+``
    in filenames will be substituted.

    ``overwrite``: If false, then don't ever overwrite anything.

    ``interactive``: If you are overwriting a file and interactive is
    true, then ask before overwriting.

    ``template_renderer``: This is a function for rendering templates (if you
    don't want to use string.Template).  It should have the signature
    ``template_renderer(content_as_string, vars_as_dict,
    filename=filename)``.
    """

    def out(msg):
        out_.write(msg)
        out_.write("\n")
        out_.flush()
        # This allows you to use a leading +dot+ in filenames which would

    # otherwise be skipped because leading dots make the file hidden:
    vars.setdefault("dot", ".")
    vars.setdefault("plus", "+")
    names = sorted(os.listdir(source))
    pad = " " * (indent * 2)
    if not os.path.exists(dest):
        if verbosity >= 1:
            out("%sCreating %s/" % (pad, dest))
        if not simulate:
            makedirs(dest, verbosity=verbosity, pad=pad)
    elif verbosity >= 2:
        out("%sDirectory %s exists" % (pad, dest))
    for name in names:
        full = os.path.join(source, name)
        reason = should_skip_file(name)
        if reason:
            if verbosity >= 2:
                reason = pad + reason % {"filename": full}
                out(reason)
            continue  # pragma: no cover
        if sub_vars:
            dest_full = os.path.join(dest, substitute_filename(name, vars))
        sub_file = False
        if dest_full.endswith("_tmpl"):
            dest_full = dest_full[:-5]
            sub_file = sub_vars
        if os.path.isdir(full):
            if verbosity:
                out("%sRecursing into %s" % (pad, os.path.basename(full)))
            copy_dir(
                full,
                dest_full,
                vars,
                verbosity,
                simulate,
                indent=indent + 1,
                sub_vars=sub_vars,
                interactive=interactive,
                template_renderer=template_renderer,
                out_=out_,
            )
            continue
        else:
            f = open(full, "rb")
            content = f.read()
            f.close()
        if sub_file:
            try:
                content = substitute_content(
                    content, vars, filename=full, template_renderer=template_renderer
                )
            except SkipTemplate:
                continue  # pragma: no cover
            if content is None:
                continue  # pragma: no cover
        already_exists = os.path.exists(dest_full)
        if already_exists:
            f = open(dest_full, "rb")
            old_content = f.read()
            f.close()
            if old_content == content:
                if verbosity:
                    out("%s%s already exists (same content)" % (pad, dest_full))
                continue  # pragma: no cover
            if interactive:
                if not query_interactive(
                    native_(full, fsenc),
                    native_(dest_full, fsenc),
                    native_(content, fsenc),
                    native_(old_content, fsenc),
                    simulate=simulate,
                    out_=out_,
                ):
                    continue
            elif not overwrite:
                continue  # pragma: no cover
        if verbosity:
            out("%sCopying %s to %s" % (pad, os.path.basename(full), dest_full))
        if not simulate:
            f = open(dest_full, "wb")
            f.write(content)
            f.close()


class SkipTemplate(Exception):
    """
    Raised to indicate that the template should not be copied over.
    Raise this exception during the substitution of your template
    """


def makedirs(dir, verbosity, pad):
    parent = os.path.dirname(os.path.abspath(dir))
    if not os.path.exists(parent):
        makedirs(parent, verbosity, pad)
    os.mkdir(dir)


def substitute_filename(fn, vars):
    for var, value in vars.items():
        fn = fn.replace("+%s+" % var, str(value))
    return fn


def substitute_content(content, vars, filename="<string>", template_renderer=None):
    v = standard_vars.copy()
    v.update(vars)
    return template_renderer(content.decode("utf-8"), v, filename=filename).encode(
        "utf-8"
    )


def should_skip_file(name):
    """
    Checks if a file should be skipped based on its name.

    If it should be skipped, returns the reason, otherwise returns
    None.
    """
    if name.startswith("."):
        return "Skipping hidden file %(filename)s"
    if name.endswith("~") or name.endswith(".bak"):
        return "Skipping backup file %(filename)s"
    if name.endswith(".pyc") or name.endswith(".pyo"):
        return "Skipping %s file " % os.path.splitext(name)[1] + "%(filename)s"
    if name.endswith("$py.class"):
        return "Skipping $py.class file %(filename)s"
    if name in ("CVS", "_darcs"):
        return "Skipping version control directory %(filename)s"
    return None


# Overridden on user's request:
all_answer = None


def query_interactive(
    src_fn, dest_fn, src_content, dest_content, simulate, out_=sys.stdout
):
    def out(msg):
        out_.write(msg)
        out_.write("\n")
        out_.flush()

    global all_answer
    from difflib import context_diff, unified_diff

    u_diff = list(
        unified_diff(
            dest_content.splitlines(), src_content.splitlines(), dest_fn, src_fn
        )
    )
    c_diff = list(
        context_diff(
            dest_content.splitlines(), src_content.splitlines(), dest_fn, src_fn
        )
    )
    added = len(
        [
            l_added
            for l_added in u_diff
            if l_added.startswith("+") and not l_added.startswith("+++")
        ]
    )
    removed = len(
        [
            l_removed
            for l_removed in u_diff
            if l_removed.startswith("-") and not l_removed.startswith("---")
        ]
    )
    if added > removed:
        msg = "; %i lines added" % (added - removed)
    elif removed > added:
        msg = "; %i lines removed" % (removed - added)
    else:
        msg = ""
    out(
        "Replace %i bytes with %i bytes (%i/%i lines changed%s)"
        % (
            len(dest_content),
            len(src_content),
            removed,
            len(dest_content.splitlines()),
            msg,
        )
    )
    prompt = "Overwrite %s [y/n/d/B/?] " % dest_fn
    while 1:
        if all_answer is None:
            response = input(prompt).strip().lower()
        else:
            response = all_answer
        if not response or response[0] == "b":
            import shutil

            new_dest_fn = dest_fn + ".bak"
            n = 0
            while os.path.exists(new_dest_fn):
                n += 1
                new_dest_fn = dest_fn + ".bak" + str(n)
            out("Backing up %s to %s" % (dest_fn, new_dest_fn))
            if not simulate:
                shutil.copyfile(dest_fn, new_dest_fn)
            return True
        elif response.startswith("all "):
            rest = response[4:].strip()
            if not rest or rest[0] not in ("y", "n", "b"):
                out(query_usage)
                continue
            response = all_answer = rest[0]
        if response[0] == "y":
            return True
        elif response[0] == "n":
            return False
        elif response == "dc":
            out("\n".join(c_diff))
        elif response[0] == "d":
            out("\n".join(u_diff))
        else:
            out(query_usage)


query_usage = """\
Responses:
  Y(es):    Overwrite the file with the new content.
  N(o):     Do not overwrite the file.
  D(iff):   Show a unified diff of the proposed changes (dc=context diff)
  B(ackup): Save the current file contents to a .bak file
            (and overwrite)
  Type "all Y/N/B" to use Y/N/B for answer to all future questions
"""

standard_vars = {
    "nothing": None,
    "empty": '""',
    "repr": repr,
    "str": str,
    "bool": bool,
    "SkipTemplate": SkipTemplate,
}



from __future__ import print_function

import re

from gearbox.command import TemplateCommand


class MakePackageCommand(TemplateCommand):
    CLEAN_PACKAGE_NAME_RE = re.compile("[^a-zA-Z0-9_]")

    def get_description(self):
        return "Creates a basic python package"

    def get_parser(self, prog_name):
        parser = super(MakePackageCommand, self).get_parser(prog_name)

        parser.add_argument(
            "-n",
            "--name",
            dest="project",
            metavar="NAME",
            required=True,
            help="Project Name",
        )

        parser.add_argument(
            "-o",
            "--output-dir",
            dest="output_dir",
            metavar="OUTPUT_DIR",
            help="Destination directory (by default the project name)",
        )

        parser.add_argument(
            "-p",
            "--package",
            dest="package",
            metavar="PACKAGE",
            help="Python Package Name",
        )

        parser.add_argument(
            "-a",
            "--author",
            dest="author",
            metavar="AUTHOR",
            default="Unknown",
            help="Name of the package author",
        )

        parser.add_argument(
            "-e",
            "--email",
            dest="author_email",
            metavar="AUTHOR_EMAIL",
            help="Email of the package author",
        )

        parser.add_argument(
            "-u", "--url", dest="url", metavar="URL", help="Project homepage"
        )

        parser.add_argument(
            "-l",
            "--license",
            dest="license_name",
            metavar="LICENSE_NAME",
            help="License used for the project",
        )

        parser.add_argument(
            "-d",
            "--description",
            dest="description",
            metavar="DESCRIPTION",
            help="Package description",
        )

        parser.add_argument(
            "-k",
            "--keywords",
            dest="keywords",
            metavar="KEYWORDS",
            help="Package keywords",
        )

        return parser

    def take_action(self, opts):
        if opts.package is None:
            opts.package = self.CLEAN_PACKAGE_NAME_RE.sub("", opts.project.lower())

        if opts.output_dir is None:
            opts.output_dir = opts.project

        opts.zip_safe = False
        opts.version = "0.0.1"

        self.run_template(opts.output_dir, opts)



import argparse
import inspect
import os
import sys

from .template import GearBoxTemplate


class Command(object):
    deprecated = False

    def __init__(self, app, app_args, cmd_name=None):
        self.app = app
        self.app_args = app_args
        self.cmd_name = cmd_name

    def get_description(self):
        """Override to provide custom description for command."""
        return inspect.getdoc(self.__class__) or ""

    def get_parser(self, prog_name):
        """Override to add command options."""
        parser = argparse.ArgumentParser(
            description=self.get_description(), prog=prog_name, add_help=False
        )
        return parser

    def take_action(self, parsed_args):
        """Override to do something useful."""
        raise NotImplementedError

    def run(self, parsed_args):
        self.take_action(parsed_args)
        return 0


class TemplateCommand(Command):
    template = GearBoxTemplate()

    def get_template_path(self):
        module = sys.modules[self.__class__.__module__]
        module_path = module.__file__
        return os.path.join(os.path.abspath(os.path.dirname(module_path)), "template")

    def run_template(self, output_dir, opts):
        self.template.run(self.get_template_path(), output_dir, vars(opts))



# -*- coding: utf-8 -*-

import argparse
import sys
import traceback

from ..command import Command


class HelpAction(argparse.Action):
    """Provide a custom action so the -h and --help options
    to the main app will print a list of the commands.
    The commands are determined by checking the CommandManager
    instance, passed in as the "default" value for the action.
    """

    def __call__(self, parser, namespace, values, option_string=None):
        app = self.default
        parser.print_help(sys.stdout)
        print("\nCommands:")
        command_manager = app.command_manager
        for name, ep in sorted(command_manager):
            try:
                factory = ep.load()
            except Exception:
                print("Could not load %r" % ep)
                if namespace.debug:
                    traceback.print_exc(file=sys.stdout)
                continue
            try:
                cmd = factory(app, None)
                if cmd.deprecated:
                    continue
            except Exception as err:
                print("Could not instantiate %r: %s\n" % (ep, err))
                if namespace.debug:
                    traceback.print_exc(file=sys.stdout)
                continue
            one_liner = cmd.get_description().split("\n")[0]
            print("  %-13s  %s" % (name, one_liner))
        sys.exit(0)


class HelpCommand(Command):
    """print detailed help for another command"""

    def get_parser(self, prog_name):
        parser = super(HelpCommand, self).get_parser(prog_name)
        parser.add_argument(
            "cmd",
            nargs="*",
            help="name of the command",
        )
        return parser

    def take_action(self, parsed_args):
        if not parsed_args.cmd:
            action = HelpAction(None, None, default=self.app)
            action(self.app.parser, self.app.parser, None, None)
            return 1

        try:
            the_cmd = self.app.command_manager.find_command(
                parsed_args.cmd,
            )
            cmd_factory, cmd_name, search_args = the_cmd
        except ValueError:
            # Did not find an exact match
            cmd = parsed_args.cmd[0]
            fuzzy_matches = [
                k[0] for k in self.app.command_manager if k[0].startswith(cmd)
            ]
            if not fuzzy_matches:
                raise
            print('Command "%s" matches:' % cmd)
            for fm in sorted(fuzzy_matches):
                print("  %s" % fm)
            return

        self.app_args.cmd = search_args
        cmd = cmd_factory(self.app, self.app_args)
        full_name = " ".join([self.app.NAME, cmd_name])
        cmd_parser = cmd.get_parser(full_name)
        cmd_parser.print_help(sys.stdout)
        return 0



from __future__ import print_function

import os
from argparse import RawDescriptionHelpFormatter

from gearbox.command import Command
from gearbox.template import GearBoxTemplate
from gearbox.utils.plugins import find_egg_info_dir


class ScaffoldCommand(Command):
    def get_description(self):
        return """Creates a new file from a scaffold template

Scaffold templates are recursively looked up inside the current path,
for example TurboGears2 projects can easily create new models, controllers
and template using something like:

    $ gearbox scaffold model controller template something

which will create a SomethingModel, SomethingController and something.html
using model/model.py.template, controllers/controller.py.template and
templates/template.html.template scaffolds of the current project.
"""

    def get_parser(self, prog_name):
        parser = super(ScaffoldCommand, self).get_parser(prog_name)

        parser.formatter_class = RawDescriptionHelpFormatter

        parser.add_argument(
            "scaffold_name", nargs="+", help="One or more scaffold templates to use"
        )

        parser.add_argument("target", help="entity for which scaffold is created")

        parser.add_argument(
            "-l",
            "--lookup",
            dest="lookup",
            help="Path where to lookup for scaffold templates.",
        )

        parser.add_argument(
            "-p",
            "--path",
            dest="path",
            help="Where to place the newly created files, "
            "by default same directory of the template",
        )

        parser.add_argument(
            "-s",
            "--subdir",
            dest="subdir",
            help="Place the newly created path in a subdir, "
            "instead of directly placing them into the path.",
        )

        parser.add_argument(
            "-np",
            "--no-package",
            dest="nopackage",
            help="When using subdir option do not create python "
            "packages, but plain directories.",
        )

        return parser

    def _lookup(self, template, where):
        template_filename = None
        for root, __, files in os.walk(where):
            for f in files:
                fname, fext = os.path.splitext(f)
                if fext == ".template" and os.path.splitext(fname)[0] == template:
                    template_filename = os.path.join(root, f)
                    break
        return template_filename

    def _find_toplevel_packages(self):
        egg_info_dir = find_egg_info_dir(os.getcwd())
        if egg_info_dir is None:
            print(
                "Unable to find distribution egg-info, did you run setup.py egg_info or develop?"
            )
            return []

        modules = os.path.join(egg_info_dir, "top_level.txt")
        try:
            modules = open(modules).readlines()
        except Exception:
            print("Unable to detect distribution top level packages.")
            return []

        modules = map(lambda x: x.strip().replace(os.sep, "."), modules)
        modules = filter(
            lambda x: not x.startswith("test"), modules
        )  # Remove test suite modules
        return list(modules)

    def take_action(self, opts):
        for template in opts.scaffold_name:
            template_filename = None
            template_relpath = None
            if template.endswith(".template"):
                # Template is a path, use it as it is.
                template_filename = template
                template_relpath = "."
            else:
                # Not a template path, look it up in subfolders
                template_filename = self._lookup(template, ".")
                if template_filename:
                    template_relpath = os.path.dirname(template_filename)

                if template_filename is None and opts.lookup:
                    template_filename = self._lookup(template, opts.lookup)
                    if template_filename:
                        template_relpath = os.path.relpath(
                            os.path.dirname(template_filename), opts.lookup
                        )

            if not template_filename or not os.path.exists(template_filename):
                print("Template %s Not Found!" % (template))
                continue

            print("Using %s for %s" % (template_filename, opts.target))
            template_with_ext, __ = os.path.splitext(template_filename)
            __, output_ext = os.path.splitext(template_with_ext)

            output_dir = opts.path
            if not output_dir:
                output_dir = template_relpath

            if opts.subdir:
                output_dir = os.path.join(output_dir, opts.subdir)

            if not os.path.exists(output_dir):
                os.makedirs(output_dir)
                if opts.subdir:
                    package_init = os.path.join(output_dir, "__init__.py")
                    if not os.path.exists(package_init):
                        with open(package_init, "w") as pif:
                            pif.write("# -*- coding: utf-8 -*-\n")

            output_path = os.path.join(output_dir, opts.target) + output_ext
            print("Creating %s..." % output_path)

            with open(template_filename, "r") as tf:
                try:
                    subdir_as_package = (
                        opts.subdir.replace(os.sep, ".") if opts.subdir else ""
                    )
                    text = GearBoxTemplate().template_renderer(
                        tf.read(),
                        {
                            "target": opts.target,
                            "subdir": opts.subdir,
                            "subpackage": subdir_as_package,
                            "dotted_subpackage": "." + subdir_as_package,
                            "packages": self._find_toplevel_packages(),
                        },
                    )
                except NameError as e:
                    print("!! Error while processing template: %s" % e)
                    continue

                with open(output_path, "w") as of:
                    of.write(text)



# (c) 2005 Ian Bicking and contributors; written for Paste
# (http://pythonpaste.org) Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license.php
#
# For discussion of daemonizing:
# http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/278731
#
# Code taken also from QP: http://www.mems-exchange.org/software/qp/ From
# lib/site.py

import atexit
import ctypes
import errno
import logging
import os
import platform
import re
import subprocess
import sys
import threading
import time

import hupper
import hupper.reloader
from paste.deploy import loadapp, loadserver
from paste.deploy.converters import asbool

from gearbox.command import Command
from gearbox.utils.log import setup_logging

MAXFD = 1024

if platform.system() == "Windows" and not hasattr(os, "kill"):  # pragma: no cover
    # py 2.6 on windows
    def kill(pid, sig=None):
        """kill function for Win32"""
        # signal is ignored, semibogus raise message
        kernel32 = ctypes.windll.kernel32
        handle = kernel32.OpenProcess(1, 0, pid)
        if 0 == kernel32.TerminateProcess(handle, 0):
            raise OSError("No such process %s" % pid)
else:
    kill = os.kill


class DaemonizeException(Exception):
    pass


class ServeCommand(Command):
    _scheme_re = re.compile(r"^[a-z][a-z]+:", re.I)

    _monitor_environ_key = "PASTE_MONITOR_SHOULD_RUN"

    possible_subcommands = ("start", "stop", "restart", "status")

    def get_parser(self, prog_name):
        parser = super(ServeCommand, self).get_parser(prog_name)

        parser.add_argument(
            "-c",
            "--config",
            help="application config file to read (default: development.ini)",
            dest="config_file",
            default="development.ini",
        )

        parser.add_argument(
            "-n",
            "--app-name",
            dest="app_name",
            metavar="NAME",
            help="Load the named application (default main)",
        )
        parser.add_argument(
            "-s",
            "--server",
            dest="server",
            metavar="SERVER_TYPE",
            help="Use the named server.",
        )
        parser.add_argument(
            "--server-name",
            dest="server_name",
            metavar="SECTION_NAME",
            help=(
                "Use the named server as defined in the configuration file "
                "(default: main)"
            ),
        )
        if hasattr(os, "fork"):
            parser.add_argument(
                "--daemon",
                dest="daemon",
                action="store_true",
                help="Run in daemon (background) mode",
            )
        parser.add_argument(
            "--pid-file",
            dest="pid_file",
            metavar="FILENAME",
            help=(
                "Save PID to file (default to gearbox.pid if running in daemon mode)"
            ),
        )
        parser.add_argument(
            "--reload",
            dest="reload",
            action="store_true",
            help="Use auto-restart file monitor",
        )
        parser.add_argument(
            "--reload-interval",
            dest="reload_interval",
            default=1,
            help=(
                "Seconds between checking files (low number can cause "
                "significant CPU usage)"
            ),
        )
        parser.add_argument(
            "--monitor-restart",
            dest="monitor_restart",
            action="store_true",
            help="Auto-restart server if it dies",
        )
        parser.add_argument(
            "--status",
            action="store_true",
            dest="show_status",
            help="Show the status of the (presumably daemonized) server",
        )

        if hasattr(os, "setuid"):
            # I don't think these are available on Windows
            parser.add_argument(
                "--user",
                dest="set_user",
                metavar="USERNAME",
                help="Set the user (usually only possible when run as root)",
            )
            parser.add_argument(
                "--group",
                dest="set_group",
                metavar="GROUP",
                help="Set the group (usually only possible when run as root)",
            )

        parser.add_argument(
            "--stop-daemon",
            dest="stop_daemon",
            action="store_true",
            help=(
                "Stop a daemonized server (given a PID file, or default "
                "gearbox.pid file)"
            ),
        )

        parser.add_argument("args", nargs="*")

        return parser

    def get_description(self):
        return "Serves a web application that uses a PasteDeploy configuration file"

    @classmethod
    def out(cls, msg, error=False):  # pragma: no cover
        log = logging.getLogger("gearbox")
        if error:
            log.error(msg)
        else:
            log.info(msg)

    def take_action(self, opts):
        if opts.stop_daemon:
            return self.stop_daemon(opts)

        if not hasattr(opts, "set_user"):
            # Windows case:
            opts.set_user = opts.set_group = None

        self.verbose = self.app_args.verbose_level

        # @@: Is this the right stage to set the user at?
        self.change_user_group(opts.set_user, opts.set_group)

        app_spec = opts.config_file
        if opts.args and opts.args[0] in self.possible_subcommands:
            cmd = opts.args[0]
            restvars = opts.args[1:]
        else:
            cmd = None
            restvars = opts.args[0:]

        if cmd not in (None, "start", "stop", "restart", "status"):
            self.out("Error: must give start|stop|restart (not %s)" % cmd)
            return 2

        if cmd == "status" or opts.show_status:
            return self.show_status(opts)

        if opts.monitor_restart and opts.reload:
            self.out("Cannot user --monitor-restart with --reload")
            return 2

        if opts.monitor_restart and not os.environ.get(self._monitor_environ_key):
            # gearbox serve was started with an angel and we are not already inside the angel.
            # Switch this process to being the angel and start a new one with the real server.
            return self.restart_with_monitor()

        if opts.reload and not hupper.is_active():
            if self.verbose > 1:
                self.out("Running reloading file monitor")
            reloader = hupper.reloader.Reloader(
                worker_path="gearbox.main.main",
                reload_interval=opts.reload_interval,
                monitor_factory=hupper.reloader.find_default_monitor_factory(
                    logging.getLogger("gearbox")
                ),
                logger=logging.getLogger("gearbox"),
            )
            reloader.run()

        if hupper.is_active():
            # Tack also config file changes
            hupper.get_reloader().watch_files([opts.config_file])

        if cmd == "restart" or cmd == "stop":
            result = self.stop_daemon(opts)
            if result:
                if cmd == "restart":
                    self.out("Could not stop daemon; aborting")
                else:
                    self.out("Could not stop daemon")
                return result
            if cmd == "stop":
                return result
            opts.daemon = True

        if cmd == "start":
            opts.daemon = True

        app_name = opts.app_name
        parsed_vars = self.parse_vars(restvars)
        if not self._scheme_re.search(app_spec):
            app_spec = "config:" + app_spec
        server_name = opts.server_name
        if opts.server:
            server_spec = "egg:gearbox"
            assert server_name is None
            server_name = opts.server
        else:
            server_spec = app_spec
        base = os.getcwd()

        if getattr(opts, "daemon", False):
            if not opts.pid_file:
                opts.pid_file = "gearbox.pid"

        # Ensure the pid file is writeable
        if opts.pid_file:
            try:
                writeable_pid_file = open(opts.pid_file, "a")
            except IOError as ioe:
                msg = "Error: Unable to write to pid file: %s" % ioe
                raise ValueError(msg)
            writeable_pid_file.close()

        if getattr(opts, "daemon", False):
            try:
                self.daemonize(opts)
            except DaemonizeException as ex:
                if self.verbose > 0:
                    self.out(str(ex))
                return 2

        if opts.pid_file:
            self.record_pid(opts.pid_file)

        log_fn = app_spec
        if log_fn.startswith("config:"):
            log_fn = app_spec[len("config:") :]
        elif log_fn.startswith("egg:"):
            log_fn = None

        if self.app.options.log_file:
            stdout_log = LazyWriter(self.app.options.log_file, "a")
            sys.stdout = stdout_log
            sys.stderr = stdout_log

        try:
            server = self.loadserver(
                server_spec, name=server_name, relative_to=base, global_conf=parsed_vars
            )
        except Exception:
            self.out("Failed to load server", error=True)
            raise

        if log_fn:
            log_fn = os.path.join(base, log_fn)
            setup_logging(log_fn)

        try:
            app = self.loadapp(
                app_spec, name=app_name, relative_to=base, global_conf=parsed_vars
            )
        except Exception:
            self.out("Failed to load application", error=True)
            raise

        if self.verbose > 0:
            if hasattr(os, "getpid"):
                msg = "Starting server in PID %i." % os.getpid()
            else:
                msg = "Starting server."
            self.out(msg)

        def serve():
            try:
                server(app)
            except (SystemExit, KeyboardInterrupt) as e:
                if self.verbose > 1:
                    raise
                if str(e):
                    msg = " " + str(e)
                else:
                    msg = ""
                self.out("Exiting%s (-v to see traceback)" % msg)

        serve()

    def loadserver(self, server_spec, name, relative_to, **kw):  # pragma:no cover
        return loadserver(server_spec, name=name, relative_to=relative_to, **kw)

    def loadapp(self, app_spec, name, relative_to, **kw):  # pragma: no cover
        return loadapp(app_spec, name=name, relative_to=relative_to, **kw)

    def parse_vars(self, args):
        """
        Given variables like ``['a=b', 'c=d']`` turns it into ``{'a':
        'b', 'c': 'd'}``
        """
        result = {}
        for arg in args:
            if "=" not in arg:
                raise ValueError('Variable assignment %r invalid (no "=")' % arg)
            name, value = arg.split("=", 1)
            result[name] = value
        return result

    def get_fixed_argv(self):  # pragma: no cover
        """Get proper arguments for re-running the command.

        This is primarily for fixing some issues under Windows.

        First, there was a bug in Windows when running an executable
        located at a path with a space in it. This has become a
        non-issue with current versions of Python and Windows,
        so we don't take measures like adding quotes or calling
        win32api.GetShortPathName() as was necessary in former times.

        Second, depending on whether gearbox was installed as an egg
        or a wheel under Windows, it is run as a .py or an .exe stub.
        In the first case, we need to run it through the interpreter.
        On other operating systems, we can re-run the command as is.

        """
        argv = sys.argv[:]
        if sys.platform == "win32" and argv[0].endswith(".py"):
            argv.insert(0, sys.executable)
        return argv

    def daemonize(self, opts):  # pragma: no cover
        pid = live_pidfile(opts.pid_file)
        if pid:
            raise DaemonizeException(
                "Daemon is already running (PID: %s from PID file %s)"
                % (pid, opts.pid_file)
            )

        if self.verbose > 0:
            self.out("Entering daemon mode")
        pid = os.fork()
        if pid:
            # The forked process also has a handle on resources, so we
            # *don't* want proper termination of the process, we just
            # want to exit quick (which os._exit() does)
            os._exit(0)
            # Make this the session leader
        os.setsid()
        # Fork again for good measure!
        pid = os.fork()
        if pid:
            os._exit(0)

        # @@: Should we set the umask and cwd now?

        import resource  # Resource usage information.

        maxfd = resource.getrlimit(resource.RLIMIT_NOFILE)[1]
        if maxfd == resource.RLIM_INFINITY:
            maxfd = MAXFD
            # Iterate through and close all file descriptors.
        for fd in range(0, maxfd):
            try:
                os.close(fd)
            except OSError:  # ERROR, fd wasn't open to begin with (ignored)
                pass

        if hasattr(os, "devnull"):
            REDIRECT_TO = os.devnull
        else:
            REDIRECT_TO = "/dev/null"
        os.open(REDIRECT_TO, os.O_RDWR)  # standard input (0)
        # Duplicate standard input to standard output and standard error.
        os.dup2(0, 1)  # standard output (1)
        os.dup2(0, 2)  # standard error (2)

    def _remove_pid_file(self, written_pid, filename, verbosity):
        current_pid = os.getpid()
        if written_pid != current_pid:
            # A forked process must be exiting, not the process that
            # wrote the PID file
            return
        if not os.path.exists(filename):
            return
        with open(filename) as f:
            content = f.read().strip()
        try:
            pid_in_file = int(content)
        except ValueError:
            pass
        else:
            if pid_in_file != current_pid:
                msg = "PID file %s contains %s, not expected PID %s"
                self.out(msg % (filename, pid_in_file, current_pid))
                return
        if verbosity > 0:
            self.out("Removing PID file %s" % filename)
        try:
            os.unlink(filename)
            return
        except OSError as e:
            # Record, but don't give traceback
            self.out("Cannot remove PID file: (%s)" % e)
            # well, at least lets not leave the invalid PID around...
        try:
            with open(filename, "w") as f:
                f.write("")
        except OSError as e:
            self.out("Stale PID left in file: %s (%s)" % (filename, e))
        else:
            self.out("Stale PID removed")

    def record_pid(self, pid_file):
        pid = os.getpid()
        if self.verbose > 1:
            self.out("Writing PID %s to %s" % (pid, pid_file))
        with open(pid_file, "w") as f:
            f.write(str(pid))
        atexit.register(self._remove_pid_file, pid, pid_file, self.verbose)

    def stop_daemon(self, opts):  # pragma: no cover
        pid_file = opts.pid_file or "gearbox.pid"
        if not os.path.exists(pid_file):
            self.out("No PID file exists in %s" % pid_file)
            return 1
        pid = read_pidfile(pid_file)
        if not pid:
            self.out("Not a valid PID file in %s" % pid_file)
            return 1
        pid = live_pidfile(pid_file)
        if not pid:
            self.out("PID in %s is not valid (deleting)" % pid_file)
            try:
                os.unlink(pid_file)
            except (OSError, IOError) as e:
                self.out("Could not delete: %s" % e)
                return 2
            return 1
        for j in range(10):
            if not live_pidfile(pid_file):
                break
            import signal

            kill(pid, signal.SIGTERM)
            time.sleep(1)
        else:
            self.out("failed to kill web process %s" % pid)
            return 3
        if os.path.exists(pid_file):
            os.unlink(pid_file)
        return 0

    def show_status(self, opts):  # pragma: no cover
        pid_file = opts.pid_file or "gearbox.pid"
        if not os.path.exists(pid_file):
            self.out("No PID file %s" % pid_file)
            return 1
        pid = read_pidfile(pid_file)
        if not pid:
            self.out("No PID in file %s" % pid_file)
            return 1
        pid = live_pidfile(pid_file)
        if not pid:
            self.out("PID %s in %s is not running" % (pid, pid_file))
            return 1
        self.out("Server running in PID %s" % pid)
        return 0

    def restart_with_monitor(self):  # pragma: no cover
        if self.verbose > 0:
            self.out("Starting subprocess with angel")

        while 1:
            args = self.get_fixed_argv()
            new_environ = os.environ.copy()
            new_environ[self._monitor_environ_key] = "true"
            proc = None
            try:
                try:
                    _turn_sigterm_into_systemexit()
                    proc = subprocess.Popen(args, env=new_environ)
                    proc = None
                except KeyboardInterrupt:
                    self.out("^C caught in monitor process")
                    if self.verbose > 1:
                        raise
                    return 1
            finally:
                if proc is not None:
                    import signal

                    try:
                        kill(proc.pid, signal.SIGTERM)
                    except (OSError, IOError):
                        pass

            if self.verbose > 0:
                self.out("%s %s %s" % ("-" * 20, "Server died, restarting", "-" * 20))

    def change_user_group(self, user, group):  # pragma: no cover
        if not user and not group:
            return
        import grp
        import pwd

        uid = gid = None
        if group:
            try:
                gid = int(group)
                group = grp.getgrgid(gid).gr_name
            except ValueError:
                import grp

                try:
                    entry = grp.getgrnam(group)
                except KeyError:
                    raise ValueError("Bad group: %r; no such group exists" % group)
                gid = entry.gr_gid
        try:
            uid = int(user)
            user = pwd.getpwuid(uid).pw_name
        except ValueError:
            try:
                entry = pwd.getpwnam(user)
            except KeyError:
                raise ValueError("Bad username: %r; no such user exists" % user)
            if not gid:
                gid = entry.pw_gid
            uid = entry.pw_uid
        if self.verbose > 0:
            self.out(
                "Changing user to %s:%s (%s:%s)"
                % (user, group or "(unknown)", uid, gid)
            )
        if gid:
            os.setgid(gid)
        if uid:
            os.setuid(uid)


class LazyWriter(object):
    """
    File-like object that opens a file lazily when it is first written
    to.
    """

    def __init__(self, filename, mode="w"):
        self.filename = filename
        self.fileobj = None
        self.lock = threading.Lock()
        self.mode = mode

    def open(self):
        if self.fileobj is None:
            with self.lock:
                self.fileobj = open(self.filename, self.mode)
        return self.fileobj

    def close(self):
        fileobj = self.fileobj
        if fileobj is not None:
            fileobj.close()

    def __del__(self):
        self.close()

    def write(self, text):
        fileobj = self.open()
        fileobj.write(text)
        fileobj.flush()

    def writelines(self, text):
        fileobj = self.open()
        fileobj.writelines(text)
        fileobj.flush()

    def flush(self):
        self.open().flush()


def live_pidfile(pidfile):  # pragma: no cover
    """(pidfile:str) -> int | None
    Returns an int found in the named file, if there is one,
    and if there is a running process with that process id.
    Return None if no such process exists.
    """
    pid = read_pidfile(pidfile)
    if pid:
        try:
            kill(int(pid), 0)
            return pid
        except OSError as e:
            if e.errno == errno.EPERM:
                return pid
    return None


def read_pidfile(filename):
    if os.path.exists(filename):
        try:
            with open(filename) as f:
                content = f.read()
            return int(content.strip())
        except (ValueError, IOError):
            return None
    else:
        return None


def _turn_sigterm_into_systemexit():  # pragma: no cover
    """
    Attempts to turn a SIGTERM exception into a SystemExit exception.
    """
    try:
        import signal
    except ImportError:
        return

    def handle_term(signo, frame):
        raise SystemExit

    signal.signal(signal.SIGTERM, handle_term)


# For paste.deploy server instantiation (egg:gearbox#wsgiref)
def wsgiref_server_runner(wsgi_app, global_conf, **kw):  # pragma: no cover
    """
    Entry point for wsgiref's WSGI server

    Additional parameters:

    ``certfile``, ``keyfile``

        Optional SSL certificate file and host key file names. You can
        generate self-signed test files as follows:

            $ openssl genrsa 1024 > keyfile
            $ chmod 400 keyfile
            $ openssl req -new -x509 -nodes -sha1 -days 365  \\
                          -key keyfile > certfile
            $ chmod 400 certfile

        The file names should contain full paths.

    """
    from wsgiref.simple_server import WSGIServer, make_server

    host = kw.get("host", "0.0.0.0")
    port = int(kw.get("port", 8080))
    threaded = asbool(kw.get("wsgiref.threaded", False))

    server_class = WSGIServer
    certfile = kw.get("wsgiref.certfile")
    keyfile = kw.get("wsgiref.keyfile")
    scheme = "http"

    if certfile and keyfile:
        """
        based on code from nullege:
        description='Dropbox REST API Client with more consistent responses.',
        author='Rick van Hattem',
        author_email='Rick@Wol.ph',
        url='http://wol.ph/',
        """
        import ssl

        class SecureWSGIServer(WSGIServer):
            def get_request(self):
                socket, client_address = WSGIServer.get_request(self)
                socket = ssl.wrap_socket(
                    socket, server_side=True, certfile=certfile, keyfile=keyfile
                )
                return socket, client_address

        port = int(kw.get("port", 4443))
        server_class = SecureWSGIServer

    if threaded:
        try:
            from socketserver import ThreadingMixIn
        except ImportError:
            from SocketServer import ThreadingMixIn

        class GearboxWSGIServer(ThreadingMixIn, server_class):
            pass

        server_type = "Threaded"
    else:

        class GearboxWSGIServer(server_class):
            pass

        server_type = "Standard"

    server = make_server(host, port, wsgi_app, server_class=GearboxWSGIServer)
    if certfile and keyfile:
        server_type += " Secure"
        scheme += "s"
    ServeCommand.out(
        "Starting %s HTTP server on %s://%s:%s" % (server_type, scheme, host, port)
    )
    server.serve_forever()


# For paste.deploy server instantiation (egg:gearbox#gevent)
def gevent_server_factory(global_config, **kw):
    from gevent import reinit

    try:
        from gevent.pywsgi import WSGIServer
    except ImportError:
        from gevent.wsgi import WSGIServer
    from gevent.monkey import patch_all

    reinit()
    patch_all(dns=False)

    host = kw.get("host", "0.0.0.0")
    port = int(kw.get("port", 8080))

    def _gevent_serve(wsgi_app):
        ServeCommand.out("Starting Gevent HTTP server on http://%s:%s" % (host, port))
        WSGIServer((host, port), wsgi_app).serve_forever()

    return _gevent_serve


# For paste.deploy server instantiation (egg:gearbox#cherrypy)
def cherrypy_server_runner(
    app,
    global_conf=None,
    host="127.0.0.1",
    port=None,
    ssl_pem=None,
    protocol_version=None,
    numthreads=None,
    server_name=None,
    max=None,
    request_queue_size=None,
    timeout=None,
):  # pragma: no cover
    """
    Entry point for CherryPy's WSGI server

    Serves the specified WSGI app via CherryPyWSGIServer.

    ``app``

        The WSGI 'application callable'; multiple WSGI applications
        may be passed as (script_name, callable) pairs.

    ``host``

        This is the ipaddress to bind to (or a hostname if your
        nameserver is properly configured).  This defaults to
        127.0.0.1, which is not a public interface.

    ``port``

        The port to run on, defaults to 8080 for HTTP, or 4443 for
        HTTPS. This can be a string or an integer value.

    ``ssl_pem``

        This an optional SSL certificate file (via OpenSSL) You can
        generate a self-signed test PEM certificate file as follows:

            $ openssl genrsa 1024 > host.key
            $ chmod 400 host.key
            $ openssl req -new -x509 -nodes -sha1 -days 365  \\
                          -key host.key > host.cert
            $ cat host.cert host.key > host.pem
            $ chmod 400 host.pem

    ``protocol_version``

        The protocol used by the server, by default ``HTTP/1.1``.

    ``numthreads``

        The number of worker threads to create.

    ``server_name``

        The string to set for WSGI's SERVER_NAME environ entry.

    ``max``

        The maximum number of queued requests. (defaults to -1 = no
        limit).

    ``request_queue_size``

        The 'backlog' argument to socket.listen(); specifies the
        maximum number of queued connections.

    ``timeout``

        The timeout in seconds for accepted connections.
    """
    is_ssl = False
    if ssl_pem:
        port = port or 4443
        is_ssl = True

    if not port:
        if ":" in host:
            host, port = host.split(":", 1)
        else:
            port = 8080
    bind_addr = (host, int(port))

    kwargs = {}
    for var_name in ("numthreads", "max", "request_queue_size", "timeout"):
        var = locals()[var_name]
        if var is not None:
            kwargs[var_name] = int(var)

    server = None
    try:
        # Try to import from newer CherryPy releases.
        import cheroot.wsgi as wsgiserver

        server = wsgiserver.Server(bind_addr, app, server_name=server_name, **kwargs)
    except ImportError:
        # Nope. Try to import from older CherryPy releases.
        # We might just take another ImportError here. Oh well.
        from cherrypy import wsgiserver

        server = wsgiserver.CherryPyWSGIServer(
            bind_addr, app, server_name=server_name, **kwargs
        )

    server.ssl_certificate = server.ssl_private_key = ssl_pem
    if protocol_version:
        server.protocol = protocol_version

    try:
        protocol = is_ssl and "https" or "http"
        if host == "0.0.0.0":
            print(
                "serving on 0.0.0.0:%s view at %s://127.0.0.1:%s"
                % (port, protocol, port)
            )
        else:
            print("serving on %s://%s:%s" % (protocol, host, port))
        server.start()
    except (KeyboardInterrupt, SystemExit):
        server.stop()

    return server



from __future__ import print_function

import os

from paste.deploy import appconfig

from gearbox.command import Command


class SetupAppCommand(Command):
    def get_description(self):
        return "Setup an application, given a config file"

    def get_parser(self, prog_name):
        parser = super(SetupAppCommand, self).get_parser(prog_name)

        parser.add_argument(
            "-c",
            "--config",
            help="application config file to read (default: development.ini)",
            dest="config_file",
            default="development.ini",
        )

        parser.add_argument(
            "--name",
            action="store",
            dest="section_name",
            default=None,
            help="The name of the section to set up (default: app:main)",
        )

        return parser

    def take_action(self, opts):
        config_spec = opts.config_file
        section = opts.section_name
        if section is None:
            if "#" in config_spec:
                config_spec, section = config_spec.split("#", 1)
            else:
                section = "main"
        if ":" not in section:
            plain_section = section
            section = "app:" + section
        else:
            plain_section = section.split(":", 1)[0]
        if not config_spec.startswith("config:"):
            config_spec = "config:" + config_spec
        if plain_section != "main":
            config_spec += "#" + plain_section
        config_file = config_spec[len("config:") :].split("#", 1)[0]
        config_file = os.path.join(os.getcwd(), config_file)
        conf = appconfig(config_spec, relative_to=os.getcwd())
        # ep_name = conf.context.entry_point_name
        # ep_group = conf.context.protocol
        dist = conf.context.distribution
        if dist is None:
            raise RuntimeError(
                "The section %r is not the application (probably a filter).  You should add #section_name, where section_name is the section that configures your application"
                % plain_section
            )

        self._setup_config(
            dist, config_file, section, {}, verbosity=self.app.options.verbose_level
        )

    def _setup_config(self, dist, filename, section, vars, verbosity):
        """
        Called to setup an application, given its configuration
        file/directory.

        The default implementation calls
        ``package.websetup.setup_config(command, filename, section,
        vars)`` or ``package.websetup.setup_app(command, config,
        vars)``

        With ``setup_app`` the ``config`` object is a dictionary with
        the extra attributes ``global_conf``, ``local_conf`` and
        ``filename``
        """
        try:
            top_level_lines = dist.read_text("top_level.txt").split("\n")
        except AttributeError:
            # Backward compatbility with older PasteDeploy
            top_level_lines = dist.get_metadata_lines("top_level.txt")
        modules = [
            line.strip()
            for line in top_level_lines
            if line.strip() and not line.strip().startswith("#")
        ]

        if not modules:
            print("No modules are listed in top_level.txt")
            print("Try reinstalling the application to regenerate that file")

        websetup_executed = False
        for mod_name in modules:
            mod_name = mod_name + ".websetup"
            try:
                mod = self._import_module(mod_name)
            except ModuleNotFoundError as e:
                print(e)
                mod = None

            if mod is None:
                continue

            if hasattr(mod, "setup_app"):
                if verbosity:
                    print("Running setup_app() from %s" % mod_name)
                websetup_executed = True
                self._call_setup_app(mod.setup_app, filename, section, vars)
            elif hasattr(mod, "setup_config"):
                if verbosity:
                    print("Running setup_config() from %s" % mod_name)
                websetup_executed = True
                mod.setup_config(None, filename, section, vars)
            else:
                print(
                    "No setup_app() or setup_config() function in %s (%s)"
                    % (mod.__name__, mod.__file__)
                )

        if not websetup_executed:
            print("No websetup found in any of the top modules")

    def _call_setup_app(self, func, filename, section, vars):
        filename = os.path.abspath(filename)
        if ":" in section:
            section = section.split(":", 1)[1]
        conf = "config:%s#%s" % (filename, section)
        conf = appconfig(conf)
        conf.filename = filename
        func(None, conf, vars)

    def _import_module(self, s):
        """
        Import a module.
        """
        mod = __import__(s)
        parts = s.split(".")
        for part in parts[1:]:
            mod = getattr(mod, part)
        return mod
